/***************************************
XWindow Support for QuickCam
by Paul Chinn <loomer@svpal.org>
Modified by Scott Laird <scott@laird.com>
 
I took a bunch of this code from the source for VGB
"Virtual GameBoy" emulator by Marat Fayzullin and
Elan Feingold
*****************/
 
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/keysym.h>
#include <unistd.h>

/** MIT Shared Memory Extension for X ************************/
#include <sys/ipc.h>
#include <sys/shm.h>
#include <X11/extensions/XShm.h>
XShmSegmentInfo SHMInfo;


/** QuickCam include files */
#include "qcam.h"
#include "qcam-os.h"
 
/** Various X-related variables ******************************/

Screen *screen;
Display *disp;
Window root,win;
XColor col;
Colormap cmap;
XImage *ximage;
XEvent ev;
GC gc;
int screen_num;
unsigned long white,black;
int xstarted=0; 
int quit=0;
int autocalibrate = 0;
int depth;
int caught_error;

/* Set a flag to exit the loop at the end */

void quitprogram(int foo)
{
  quit=1;
}
 
int
error_catcher(Display * d, XErrorEvent * xeev)
{
  ++caught_error;
  return 0;
}

/** Initialize xwindows, and prepare a shared memory buffer for
 the image.  Returns pointer to shared memory buffer. */
 
char *InitXWindows(struct qcam *q)
{
  XGCValues values;
  XSizeHints hints;
  XWMHints wmhints;
  int width, height;
  char *sbuf;
  char *window_name="QuickCam";
  char *icon_name="QuickCam";
  XTextProperty windowName, iconName;
  int (*old_handler)();

  width=q->width/q->transfer_scale;
  height=q->height/q->transfer_scale;

  disp=XOpenDisplay(NULL);
  if(!disp) {printf("open display failed\n"); return NULL;}
  
  screen=DefaultScreenOfDisplay(disp);
  screen_num=DefaultScreen(disp);
  white=XWhitePixel(disp,screen_num);
  black=XBlackPixel(disp,screen_num);
  
  root=DefaultRootWindow(disp);
  
  win=XCreateSimpleWindow(disp,root,0,0,width,height,0,white,black);
  if(!win) {  
    printf("create window failed\n");
    return(NULL); 
  }
  
  /* tell window manager about our window */
  hints.flags=PSize|PMaxSize|PMinSize;
  hints.min_width=hints.max_width=hints.width=width;
  hints.min_height=hints.max_height=hints.height=height;
  wmhints.input=True;
  wmhints.flags=InputHint;

  XStringListToTextProperty(&window_name, 1 ,&windowName);
  XStringListToTextProperty(&icon_name, 1 ,&iconName);


  XSetWMProperties(disp, win, 
		   &windowName, &iconName,
		   NULL, 0,
		   &hints, &wmhints, NULL);
  
  /*  XStoreName(disp, win, "QuickCam"); */
  XSelectInput(disp, win, ExposureMask);
  XMapRaised(disp, win);
  XNextEvent(disp, &ev);
  
  gc = XCreateGC(disp, win, 0, &values);

  /*
   * We want to know when the shared memory extension is unavailable
   * (either because the X server doesn't support it or because the
   * client and the server do not run on the same machine).  To detect
   * this, we simply install our own error handler in which we count
   * the errors that occur.  If any errors occur during shared memory
   * image creation, we fall back to regular images.
   *
   * There must be a better way to do this, no???
   */
  old_handler = XSetErrorHandler(error_catcher);
  XSync(disp, 0);

  depth = DefaultDepthOfScreen(screen);
  ximage = XShmCreateImage(disp, DefaultVisual(disp, screen_num), 
			   depth, ZPixmap, NULL, &SHMInfo, width, height);
  if(!ximage) {
    printf("CreateImage Failed\n");
    return(NULL);
  }
 
  SHMInfo.shmid=shmget(IPC_PRIVATE, 
		       ximage->bytes_per_line*ximage->height,
		       IPC_CREAT|0777);

  if(SHMInfo.shmid < 0) {
    perror("shmget failed:");
    return (NULL);
  }
 
  sbuf = ximage->data = SHMInfo.shmaddr = shmat(SHMInfo.shmid, 0, 0);

  XShmAttach(disp, &SHMInfo);

  XSync(disp, 0);
  XSetErrorHandler(old_handler);

  if (caught_error) {
    fprintf(stderr, "Shared memory unavailable, using regular images\n");
    shmdt(SHMInfo.shmaddr);
    SHMInfo.shmaddr = 0;

    sbuf = malloc(((depth + 7) / 8) * width * height);
    ximage = XCreateImage(disp, DefaultVisual(disp, screen_num), 
			  depth, ZPixmap, 0, sbuf, width, height, depth,
			  width * ((depth + 7) / 8));
    if(!ximage) {
      printf("CreateImage Failed\n");
      return(NULL);
    }
  }
  signal(SIGHUP, quitprogram); 
  signal(SIGINT, quitprogram);
  signal(SIGQUIT, quitprogram); 
  signal(SIGTERM, quitprogram);
  xstarted=1;
  return(sbuf);
}


void ExitXWindows(void)
{
  if(xstarted) {
    if (ximage)
      XDestroyImage(ximage);
    XShmDetach(disp, &SHMInfo);
    if(SHMInfo.shmaddr)
      shmdt(SHMInfo.shmaddr);
    if(SHMInfo.shmid > 0)
      shmctl(SHMInfo.shmid, IPC_RMID, 0);
  }
}


int *xqc_createpalette(Colormap cmap, struct qcam *q)
{
  int *pal;
  int i, size;

  switch (q->bpp) {
    case 4:
    case 6: size=64; break;
    case 24:
    case 32: size=256; break;
    default: size=0;
  }

  pal=malloc(sizeof(int[size]));
  
  for(i=0; i<size; i++) {
    col.red =col.green = col.blue = i * (65536/size);
    if (!XAllocColor(disp, cmap, &col)) {
      fprintf(stderr,"XAllocColor failed on %d\n",i);
    }
    pal[i] = col.pixel;
  }
  return pal;
}


void xqc_sync(struct qcam *q, int *colortable, char *sbuf)
{
  int i;
  int limit;
  scanbuf *scan;
  int temp, width, height;

  qc_set(q);
  if (autocalibrate && q->cam_version == 0x01) qc_calibrate(q);

  scan=qc_scan(q);

  limit = q->height / q->transfer_scale;
  limit *= q->width / q->transfer_scale;

  for(i=0;i<limit;i++)
    {
      unsigned char r, g, b, grey;
 
      switch(q->bpp) 
	{
	case 4: 
	  r = g = b = grey = colortable[(scan[i]<<2)+1];
	  break;
	case 6:
	  r = g = b = grey = colortable[scan[i]];
	  break;
	case 24:
	  r = scan[i*3];
	  g = scan[i*3+1];
	  b = scan[i*3+2];
	  grey = (char)(((int)r+(int)g+(int)b)/3);
	}
 
      switch(depth)
	{
	case 8:			/* works - but should dither in color mode */
	  sbuf[i]=colortable[grey];
	  break;
	case 16:		/* works - XFree 3.1.2G 16bpp */
	  temp = scan[i*3+2] >> 3;
	  temp |= (int)(scan[i*3+1] >> 2) << 5;
	  temp |= (int)(scan[i*3] >> 3) << 11;
	  sbuf[2*i+1]=temp >> 8;
	  sbuf[2*i]=temp & 0xff;
	  break;
	case 24:		/* byte ordering?? */
	  sbuf[3*i] = r;
	  sbuf[3*i+1] = g;
	  sbuf[3*i+2] = b;
	  break;
	case 32:		/* byte ordering?? */
	  sbuf[(i>>2)]   = r;
	  sbuf[(i>>2)+1] = g;
	  sbuf[(i>>2)+2] = b;
	  sbuf[(i>>2)+3] = 0;
	  break;
	}
    }
 
  free(scan);
  
  width = q->width/q->transfer_scale;
  height = q->height/q->transfer_scale;
  if (SHMInfo.shmaddr) {
    XShmPutImage(disp, win, gc, ximage, 0, 0, 0, 0, width, height, False);
  } else {
    XPutImage(disp, win, gc, ximage,  0, 0, 0, 0, width, height);
  }
  XFlush(disp);
}


void usage(void)
{
  fprintf(stderr,"Usage:\n");
  fprintf(stderr,"  xqcam [options]\n");
  fprintf(stderr,"    Options:\n");
  fprintf(stderr,"      -x width   Set width\n");
  fprintf(stderr,"      -y height  Set height\n");
  fprintf(stderr,"      -p port    Set port\n");
  fprintf(stderr,"      -B bpp     Set bits per pixel\n");

  fprintf(stderr,"      -c val     Set contrast\n");
  fprintf(stderr,"      -w val     Set white balance\n");
  fprintf(stderr,"      -W         Auto-set white balance\n");
  fprintf(stderr,"      -b val     Set brightness\n");
  fprintf(stderr,"      -s val     Set scaling factor (1, 2, or 4)\n");
  fprintf(stderr,"      -t val     Set top line of scan\n");
  fprintf(stderr,"      -l val     Set left column of scan\n");
  fprintf(stderr,"      -V         Show version information\n");
  fprintf(stderr,"      -v         Verbose output\n");
  fprintf(stderr,"      -C         Use private colormap\n");
  fprintf(stderr,"      -r         Release the lock after each scan\n");
  fprintf(stderr,"      -u         Force unidirectional mode\n");

}


int main(int argc, char **argv)
{
  int arg;
  extern char *optarg;
  struct qcam *q;
  int *colortable;
  char *sbuf;
  int verbose=0;
  int privatecmap=0;
  int release_lock=0;
  Colormap cmap;
  struct timeval tv1, tv2;
  double framerate=0,fr;

  colortable=malloc(sizeof(int[256]));

  if(geteuid()) {
    fprintf(stderr,"%s: Must be installed SUID or run as root.  Exiting.\n",
	    argv[0]);
    exit(1);
  }

  q=qc_init();
  qc_initfile(q, 0);

  /* Read command line */

  while((arg=getopt(argc,argv,"WhCvx:y:p:b:B:c:d:t:l:s:S:w:Vru"))>0) { 
    switch (arg) {
    case 'x':
      q->width=atoi(optarg);
      break;
    case 'y':
      q->height=atoi(optarg);
      break;
    case 'p':
      if (!getuid())
	q->port=strtol(optarg,NULL,0);
      break;
    case 'B':
      q->bpp=atoi(optarg);
      break;
    case 'b':
      q->brightness=atoi(optarg);
      break;
    case 'c':
      q->contrast=atoi(optarg);
      break;
    case 'd':
      qc_setblacklevel(q, atoi(optarg));
      break;
    case 'w':
      q->whitebal=atoi(optarg);
      break;
    case 'V':
      fprintf(stderr,"%s: Version 0.6\n",argv[0]);
      exit(0);
      break;
    case 'h':
      usage();
      exit(0);
      break;
    case 'C':
      privatecmap=1;
      break;
    case 'v':
      verbose=1;
      break;
    case 's':
      if (qc_settransfer_scale(q, atoi(optarg)))
	fprintf(stderr, "Bad scaling factor, valid values are 1, 2, or 4\n");
      break;
    case 'S':
      qc_setsaturation(q, atoi(optarg));
      break;
    case 't':
      if (qc_settop(q, atoi(optarg)))
	fprintf(stderr, "Bad top line, valid values are 1 - 243\n");
      break;
    case 'l':
      if (qc_setleft(q, atoi(optarg)))
	fprintf(stderr, "Bad left column, valid values are 2 - 336, and must be even\n");
      break;
    case 'W':
      autocalibrate = 1;
      break;
    case 'r':
      release_lock=1;
      break;
    case 'u':
      qc_forceunidir(q);
      break;
    default:
      fprintf(stderr,"%s: Unknown option or error in option\n",argv[0]);
      usage();
      exit(1);
      break;
    }
  }

  switch (q->bpp) {
  case 4: 
  case 6:
  case 24: break;
  default:
    fprintf(stderr,"%s: Error: Unsupported bit depth\n",argv[0]);
    exit(1);
  }

  if (q->width % (20 * q->transfer_scale) != 0) {
    fprintf(stderr, "Warning: width must be multiple of 20*transfer_scale.\n");
  }

  /* Attempt to get permission to access IO ports.  Must be root */

  if (qc_open(q)) {
    fprintf(stderr,"Cannot open QuickCam; exiting.\n");
    exit(1);
  }

  if (!release_lock) {
     /* we no longer need root permission cuz we've done our last qc_open... */
     setuid(getuid());
  }

  fprintf(stderr,"Scanning from QuickCam at 0x%x at %dx%d (%d:1) @ %dbpp\n",
	  q->port,q->width,q->height,q->transfer_scale,q->bpp);

  /* Start X Display */

  if ((sbuf=InitXWindows(q))==NULL) {
    fprintf(stderr,"InitXWindows failed, exiting\n");
    exit(1);
  }

  if (privatecmap) {
    cmap = XCreateColormap(disp, win, DefaultVisual(disp, screen_num), 
			   AllocNone);
    XSetWindowColormap(disp, win, cmap);
  } else {
    cmap=DefaultColormap(disp, screen_num);
  }
  colortable=xqc_createpalette(cmap, q);


  /* Scan one image */
  if(verbose) {
    gettimeofday(&tv1,NULL);
  }
  
  if (release_lock) {
     /* Release IO privileges */
     qc_close(q);
  }

  while(!quit) {
     if (release_lock) {
        if (qc_open(q)) {
          fprintf(stderr,"Cannot open QuickCam; exiting.\n");
          exit(1);
        }
     }

    xqc_sync(q,colortable,sbuf);

     if (release_lock) {
        /* Release IO privileges */
        qc_close(q);
     }

    /* Calculate frame rate */
    if(verbose) {
      gettimeofday(&tv2,NULL);

      /* The frame rate is calculated using the TCP RTT algorithm */

      fr=(1.0/(tv2.tv_sec-tv1.tv_sec+(tv2.tv_usec-tv1.tv_usec)/1000000.0));
      if(framerate!=0)
	framerate=0.9*framerate+0.1*fr;
      else
	framerate=fr;

      fprintf(stderr,"Frame rate: %f fps       \r",framerate);
      tv1.tv_sec=tv2.tv_sec;
      tv1.tv_usec=tv2.tv_usec;
    }

  }

  ExitXWindows();

  if (!release_lock) {
     /* Release IO privileges */
     qc_close(q);
  }

  return 0;
}
